#include <sys/file.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include "config.h"
#include <utmp.h>

extern char *sys_errlist[];
extern int errno;
extern int max_users;
extern char *getpass();
extern char *crypt();
extern char *index();
extern char *ttyname();
extern char *terminaltype;

#define IN 1
#define OUT 0

char *new_gets(buffer,size,fp)
     char *buffer;
     int size;
     FILE *fp;
{
    char c;
    
    if (fgets(buffer,size, fp) == NULL)
      return(NULL);
    c = *(buffer + strlen(buffer) - 1);
    if ((c == '\n') || (c == '\r'))
      *(buffer + strlen(buffer) - 1) = 0;
    c = *(buffer + strlen(buffer) - 1);
    if ((c == '\n') || (c == '\r'))
      *(buffer + strlen(buffer) - 1) = 0;
    return(buffer);
}

#ifdef LOG_FILE
void log(which, name, host, tty)
int which;
char *name,
    *tty,
    *host;
{
    FILE *fp;
    long t;

    if (access(LOG_FILE,F_OK))
	return;
    if (fp = fopen(LOG_FILE, "a")) {
	t = time(0);
	if (which) {
	    if (name)
		fprintf(fp, "%s by %s@%s on %s at %.16s\n", which ? "Login" : "Logout", name, host, tty + 5, ctime(&t));
	    else
		fprintf(fp, "%s by %s on %s at %.16s\n", which ? "Login" : "Logout", host, tty + 5, ctime(&t));
	} else
	    fprintf(fp, "Connection closed from %s on %s at %.16s\n", host, tty + 5, ctime(&t));
	fclose(fp);
    }
}
#endif LOG_FILE

void modify_utmp(slot,name,tty,host)
     int slot;
     char *name,*tty,*host;
{
    struct utmp utmp;
    int f;

    time(&utmp.ut_time);
    strncpy(utmp.ut_line, tty,sizeof(utmp.ut_line));
    strncpy(utmp.ut_name,name,sizeof(utmp.ut_name));
    strncpy(utmp.ut_host,host,sizeof(utmp.ut_host));
    if (slot > 0 && (f = open("/etc/utmp", O_WRONLY)) >= 0) {
	lseek(f, (long) (slot * sizeof(utmp)), 0);
	write(f, (char *) &utmp, sizeof(utmp));
	close(f);
    }
    if ((f = open("/usr/adm/wtmp", O_WRONLY | O_APPEND)) >= 0) {
	write(f, (char *) &utmp, sizeof(utmp));
	close(f);
    }
}

int check_user_limits()
{
    struct utmp utmp;
    int f,
        cnt = 0;

    if (max_users) {
	if ((f = open("/etc/utmp", O_RDONLY)) >= 0) {
	    while (read(f, &utmp, sizeof(utmp))) {
		if (*(utmp.ut_line) && (strncmp(DAEMON_NAME, utmp.ut_name, sizeof(utmp.ut_name)) == 0))
		    cnt++;
	    }
	    close(f);
	}
	return (cnt > max_users);
    } else
	return (0);
}

#ifdef PASSWD_FILE
#ifdef USE_LOG
void log_use(name)
char *name;
{
    FILE *fp;
    long pos;
    int len,
        cnt = 0;
    char buffer[BUFSIZ + 1];

    if (access(USE_LOG,F_OK))
	return;
    if (fp = fopen(USE_LOG, "r+")) {
	pos = 0L;
	len = strlen(name);
	while (fgets(buffer, BUFSIZ, fp)) {
	    if (strncmp(buffer, name, len) == 0) {
		char *ptr;

		if (ptr = index(buffer, ' ')) {
		    cnt = atoi(ptr + 1);
		    fseek(fp, pos, 0);
		    break;
		}
	    }
	    pos = ftell(fp);
	}
	fprintf(fp, "%-9s %8d\n", name, cnt + 1);
	fclose(fp);
    }
}
#endif USE_LOG

struct passwd {			/* see getpwent(3) */
    char *pw_name;
    char *pw_passwd;
    int pw_uid;
    int pw_gid;
    int pw_quota;
    char *pw_comment;
    char *pw_gecos;
    char *pw_dir;
    char *pw_shell;
};

static char passwd_file[BUFSIZ + 1];
static char new_passwd_file[BUFSIZ + 1];

static void setpwfile(name)
char *name;
{
    strncpy(passwd_file, name, BUFSIZ);
    strncpy(new_passwd_file, name, BUFSIZ);
    strcat(new_passwd_file, ".new");
}

static struct passwd *getpwnam(name)
char *name;
{
    FILE *fp;
    static char buffer[BUFSIZ + 1];
    char *ptr,
        *word;
    static struct passwd entry;

    if ((fp = fopen(passwd_file, "r")) == NULL)
	return (NULL);
    while (new_gets(buffer, BUFSIZ, fp)) {
	word = buffer;
	ptr = index(word, ':');
	*(ptr++) = NULL;
	if (strcmp(word, name) == 0) {
	    entry.pw_name = word;
	    word = ptr;
	    ptr = index(word, ':');
	    *(ptr++) = NULL;
	    entry.pw_passwd = word;

	    word = ptr;
	    ptr = index(word, ':');
	    *(ptr++) = NULL;
	    entry.pw_uid = atoi(word);

	    entry.pw_gecos = ptr;
	    return (&entry);
	}
    }
    return (NULL);
}

/* chpwd: changes the password for the user with name username to passwd */
void chpwd(username, passwd)
char *username,
    *passwd;
{
    int len;
    FILE *pwd,
        *pwd_new;
    char line[BUFSIZ + 1],
         new_line[BUFSIZ + 1],
        *ptr;

    if ((pwd = fopen(passwd_file, "r")) == NULL) {
	netwrite("Couldn't open password file!\nPassword change failed!");
	return;
    }
    if ((pwd_new = fopen(new_passwd_file, "w")) == NULL) {
	netwrite("Couldn't write new new password file1\nPassword change failed!");
	fclose(pwd);
	return;
    }
    fchmod(fileno(pwd_new), 0644);
    strcat(username, ":");
    len = strlen(username);
    while (fgets(line, BUFSIZ, pwd)) {
	if (strncmp(username, line, len) == 0) {
	    ptr = index(line + len, ':');
	    sprintf(new_line, "%s%s%s", username, passwd, ptr);
	    fputs(new_line, pwd_new);
	    break;
	} else
	    fputs(line, pwd_new);
    }
    while (fgets(line, BUFSIZ, pwd))
	fputs(line, pwd_new);
    fclose(pwd);
    fclose(pwd_new);
    if (rename(new_passwd_file, passwd_file)) {
	netwrite("Error while updating password file!\nPassword change failed!");
	return;
    }
    chmod(passwd_file, 0644);
    netwrite("Password change successful\n");
}

int login(name)
char *name;
{
    char *pass,
        *enc,
         salt[3];
    struct passwd *entry;

    pass = getpass("Password:");
    setpwfile(PASSWD_FILE);
    if (entry = getpwnam(name)) {
	if (strlen(entry->pw_passwd)) {
	    salt[0] = entry->pw_passwd[0];
	    salt[1] = entry->pw_passwd[1];
	    salt[2] = NULL;
	    enc = crypt(pass, salt);
	    return (strcmp(enc, entry->pw_passwd) == 0);
	} else
	    return (1);
    } else
	return (0);
}

void change_password(username)
char *username;
{
    struct passwd *pwent;
    char entered_passwd[14],
         salt[3];

    setpwfile(PASSWD_FILE);
    pwent = getpwnam(username);
    sprintf(salt, "%02x", pwent->pw_uid);
    strcpy(entered_passwd, crypt(getpass("Old password: "), salt));
    if ((*pwent->pw_passwd == '\0') || (strcmp(entered_passwd, pwent->pw_passwd) == 0)) {
	strcpy(entered_passwd, crypt(getpass("New Password: "), salt));
	if (strcmp(entered_passwd, crypt(getpass("Repeat: "), salt)) == 0) {
	    chpwd(pwent->pw_name, entered_passwd);
	    return;
	}
    }
    netwrite("Incorrect Password!\n");
}
#endif PASSWD_FILE

void motd(file)
char *file;
{
    FILE *fp;
    char buffer[BUFSIZ + 1];

    if (fp = fopen(file, "r")) {
	while (fgets(buffer, BUFSIZ, fp))
	    netwrite(buffer);
	fclose(fp);
    }
}


void start_irc(nick, name, file)
char *nick,
    *name,
    *file;
{
    char *envinit[10];
    char *args[10];
    char ircname[BUFSIZ + 1];
    char nickname[BUFSIZ + 1];
    char term[BUFSIZ + 1];
    char c;

    sprintf(ircname, "IRCNAME=%s", name);
    envinit[0] = ircname;

    sprintf(nickname, "IRCNICK=%s", nick);
    envinit[1] = nickname;

    strcpy(term, "TERM=");
    if (*(terminaltype+5) && strcmp(terminaltype+5,"unknown")){
	netwrite("Your terminal appears to be: ");
	netwrite(terminaltype+5);
	netwrite("\n");
	netwrite("Hit return to select default terminal,\nor enter your terminal type,\nor enter \"none\" for no terminal: ");
	new_gets(term + 5, BUFSIZ - 5, stdin);
    }else{
	do{
	    netwrite("Enter your terminal type,\nor enter \"none\" for no terminal: ");
	    new_gets(term + 5, BUFSIZ - 5, stdin);
	}while(*(term+5) == NULL);
    }	
    envinit[2] = term;

#ifdef IRCSERVER
    envinit[3] = IRCSERVER;
    envinit[4] = NULL;
#else
    envinit[3] = NULL;
#endif

    setuid(DAEMON_UID);
    c = 0;
    args[c++] = "irc";
    args[c++] = "-s";
    if (*(term+5) == NULL)
      envinit[2] = terminaltype;
    else{
	if (strcmp(term,"TERM=none") == 0){
	    struct sgttyb oldb;
	    char buffer[10];
	    int tty_des, mask;
	    
	    tty_des = open("/dev/tty",O_RDWR);
	    ioctl(tty_des, TIOCGETP, &oldb);
	    netwrite("Please choose your backspace key:\n0) Backspace key (^H)\n1) Delete key (^?)\nChoice: ");
	    new_gets(buffer,9,stdin);
	    switch (atoi(buffer)){
		case 1:
		  oldb.sg_erase = '\177';
		  break;
		default:
		  oldb.sg_erase = '\010';
		  break;
	    }
	    ioctl(tty_des, TIOCSETP, &oldb);
	    mask = LCRTERA | LCRTKIL | LCRTBS | LCTLECH;
	    ioctl(tty_des, TIOCLBIS, &mask);
	    args[c++] = "-d";
	}
    }
    if (file) {
	args[c++] = "-l";
	args[c++] = file;
    }
    args[c++] = NULL;

    execve(IRCII_PATH, args, envinit);
    sprintf(term, "Unable to start IRCII at this time: %s.\n", sys_errlist[errno]);
    netwrite(term);
    exit(1);
}

void irc_loop(host)
char *host;
{
    char name[BUFSIZ + 1],
         choice[BUFSIZ + 1];
    struct passwd *entry;

    if (check_user_limits()){
	netwrite("Too many people signed on right now, please try again later.\n");
	return;
    }
#ifdef BANNER_FILE
    motd(BANNER_FILE);
#endif BANNER_FILE
#ifdef PASSWD_FILE
    while (1) {
	netwrite("Nickname: ");
	if (new_gets(name, BUFSIZ, stdin) == NULL)
	    return;
	if (*name) {
	    if (login(name)) {
#ifdef LOG_FILE
		log(1, name, host, ttyname(0));
#endif LOG_FILE
#ifdef USE_LOG
		log_use(name);
#endif USE_LOG
		modify_utmp(ttyslot(),DAEMON_NAME,ttyname(0)+5,host);
#ifdef MOTD_FILE
		motd(MOTD_FILE);
#endif MOTD_FILE
		while (1) {
		    netwrite("Choose one of the following:\n0) Start IRCII\n1) Change Password\n2) Close connection.\n\nChoice: ");
		    if (fgets(choice, BUFSIZ, stdin) == NULL)
			return;
		    switch (atoi(choice)) {
			case 1:
			    change_password(name);
			    break;
			case 2:
			    return;
			default:
			    entry = getpwnam(name);
#ifdef SAVE_DIR
			    sprintf(choice,"%s/%s",SAVE_DIR,entry->pw_name);
			    start_irc(name, entry->pw_gecos, choice);
#else
			    start_irc(name, entry->pw_gecos, NULL);
#endif SAVE_DIR
			    break;
		    }
		}
	    } else
		netwrite("Login incorrect\n");
	}
    }
#else PASSWD_FILE
    while (1) {
	netwrite("Nickname: ");
	if (new_gets(name, BUFSIZ, stdin) == NULL)
	    return;
	if (*name) {
#ifdef LOG_FILE
	    log(1, name, host, ttyname(0));
#endif LOG_FILE
	    modify_utmp(ttyslot(),DAEMON_NAME,ttyname(0)+5,host);
	    start_irc(name, host, NULL);
	}
    }
#endif PASSWD_FILE
}
